/* ----------------------------------------------------------------------------
    NSTX -- tunneling network-packets over DNS

     (C) 2000 by Julien Oster and Florian Heinz

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

  -------------------------------------------------------------------------- */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <fcntl.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <sys/time.h>

#include <netinet/in.h>
#include <arpa/inet.h>
#include <arpa/nameser.h>

#include <resolv.h>
#include <netdb.h>

#include <time.h>

#include "nstx.h"
#include "nstx_dns.h"
#include "nstx_pstack.h"

static int nstxc_handle_reply(char *, int);
static int nstxc_send_packet(char *, unsigned short, char *, int);
static void nstx_dumppacket (char *, int);

static int tapfd;
static int gotpacket = 0;

int main (int argc, char * argv[]) {
  int sfd, i, j, repoll, sendnull, shortdelay, middelay;
  char readbuf[NSTX_MAXPACKET];
  char * domainname;
  struct sockaddr_in sa, ma;
  fd_set fds;
  struct timeval pollt;

  unsigned int id = time(NULL);

  if (argc < 3) {
    fprintf(stderr, "Usage: ./nstxcd <domainname> <dns-server>\n");
    fprintf(stderr, "Example: ./nstxcd tun.yomama.com 125.23.53.12\n");
    exit (3);
  }

  if (*(argv[1]) == '.') {
    domainname = strdup (argv[1]);
  } else {
    domainname = malloc (strlen (argv[1]) + 2);
    *domainname = '.';
    strcpy (domainname + 1, argv[1]);
  }

  ma.sin_family = AF_INET;
  ma.sin_addr.s_addr = INADDR_ANY;
  ma.sin_port = htons(54);

  sa.sin_family = AF_INET;
  sa.sin_addr.s_addr = inet_addr (argv[2]);
  sa.sin_port = htons(53);

  init_pstack (SENDLEN);

  sfd = nstxc_dns_init ((struct sockaddr *)&sa, sizeof(sa),
			(struct sockaddr *)&ma, sizeof(ma));

  tapfd = open("/dev/tap0", O_RDWR);
  if (tapfd < 0) {
    perror("tap-open");
    exit(1);
  }


  if (sfd < 0) {
    perror ("nstxc_dns_init()");
    exit (2);
  }

  repoll = 0;

  pollt.tv_sec = 1;
  pollt.tv_usec = 0;

  sendnull = 0;
  shortdelay = 0;
  middelay = 0;

  for (;;) {

    repoll = 0;
    sendnull = 0;
    id++;
    FD_ZERO(&fds);
    FD_SET(sfd, &fds);
    FD_SET(tapfd, &fds);

    select (tapfd+1, &fds, NULL, NULL, &pollt);

    if (FD_ISSET(sfd, &fds)) {
      j = read (sfd, readbuf, 512);
      repoll = nstxc_handle_reply (readbuf, j);
    } else
      sendnull = 1;

    if (FD_ISSET(tapfd, &fds)) {
      sendnull = 0;
      shortdelay = 20;
      i = read (tapfd, readbuf, NSTX_MAXPACKET);
      nstxc_send_packet (domainname, id, readbuf, i);
      repoll = 1;
    }

    if (sendnull)
      nstxc_send_packet (domainname, id, readbuf, 0);

    if (repoll) {
      pollt.tv_sec = 0;
      pollt.tv_usec = 0;
    } else {
      if (shortdelay) {
	fflush (stderr);
	if (! --shortdelay)
	  middelay = 20;
	pollt.tv_sec = 0;
	pollt.tv_usec = 100000;
      }
      else if (middelay) {
	--middelay;
	pollt.tv_sec = 0;
	pollt.tv_usec = 750000;
      }
      else {
	pollt.tv_sec = 2;
	pollt.tv_usec = 0;
      }
    }
    usleep (5000);
  }

  return 0;
}

static int nstxc_handle_reply (char * reply, int len) {
  char buf[255];
  unsigned short sh, compressed_offset;
  int i;
  unsigned char j;
  int repoll = 0;

  char * p = reply;
  struct dnsqueryhdr * dh = (struct dnsqueryhdr *)reply;

  struct dnsanswer da;

  struct nstxhdr * nh;
  struct nstxhdr ph;

  if (dh->rcode)
    return -16;

  if (len < sizeof(struct dnsqueryhdr))
    return -1;

  if (!ntohs(dh->ancount))
    return 0;

  p += sizeof (struct dnsqueryhdr);


  for (sh = 0; sh < ntohs(dh->qdcount); sh++) {
    *buf = 0;

    compressed_offset = ntohs(*(unsigned short*)p);
    if (compressed_offset & 0xc000) {
      compressed_offset = compressed_offset & 0x3fff;
      i = 2;
      nstx_name_to_string (buf, reply + compressed_offset, 255);
    } else {
      i = nstx_name_to_string (buf, p, 255);
    }

    if (i < 0)
      return -2;

    p += i;
    len -= i;

    p += 4;
    len -= 4;
  }

  for (sh = 0; sh < ntohs(dh->ancount); sh++) {
    *buf = 0;

    compressed_offset = ntohs(*(unsigned short*)p);
    if (compressed_offset & 0xc000) {
      compressed_offset = compressed_offset & 0x3fff;
      i = 2;
      nstx_name_to_string (buf, reply + compressed_offset, 255);
    } else {
      i = nstx_name_to_string (buf, p, 255);
    }

    if (i < 0)
      return -3;
    p += i;
    len -= i;

    if (len < sizeof (struct dnsanswer))
      return -4;

    memcpy ((char*)&da, p, sizeof(struct dnsanswer));
    i = sizeof(struct dnsanswer);
    p += i;
    len -= i;

    /* DNS SUCKS! */
    /* we must overread rdlength here, we can't just use the structure because */
    /* of padding */

    p += 2;
    len -= 2;

    if (len < sizeof (struct nstxhdr))
      return -5;

    j = *p;

    p++;
    len--;

    nh = (struct nstxhdr *) p;
    memcpy ((char*)&ph, p, sizeof (struct nstxhdr));
    nh = &ph;

    repoll = nh->flags & NSTX_MF;

    if (j - sizeof(struct nstxhdr)) {
      gotpacket = 1;
      nstx_handlepacket (p, j, &nstx_dumppacket);
    } else
      gotpacket = 0;
  }

  return repoll;
}
  
static int nstxc_send_packet (char * suffix, unsigned short id,
			      char * buf, int len) {
  char * p;
  int chunks, rest;
  char hbuf[HCHUNKLEN];
  char cbuf[FQDN_MAXLEN];
  int i;

  struct nstxhdr nh;

  nh.magic = NSTX_MAGIC;
  nh.seq = 0;
  nh.frc = 0;
  nh.id = id;
  nh.crop = 0;
  nh.flags = 2;

  p = buf;
  chunks = len / CHUNKLEN;
  rest = len % CHUNKLEN;
  nh.frc = chunks - (rest ? 0 : 1);

  while (chunks) {
    memcpy (hbuf, (char*)&nh, sizeof(nh));
    memcpy (hbuf + sizeof(nh), p, CHUNKLEN);

    nstx_build_fqdn (suffix, cbuf, hbuf, HCHUNKLEN);

    i = nstxc_dns_query (cbuf, id);
    if (i < 0)
      return -1;

    p += CHUNKLEN;
    chunks--;
    nh.seq++;
  }

  if (rest | !chunks) {
    nh.crop = 2 - (rest % 3);
    memcpy (hbuf, (char*)&nh, sizeof(nh));
    memcpy (hbuf + sizeof(nh), p, rest);

    nstx_build_fqdn (suffix, cbuf, hbuf, rest + sizeof(nh));

    i = nstxc_dns_query (cbuf, id);
    if (i < 0)
      return -2;
  }

  return 0;
}

void nstx_dumppacket(char *ptr, int len) {
   write(tapfd, ptr, len);
   return;
}
